/*
 * Copyright (C) 1997 and 1998 WIDE Project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by WIDE Project and
 *	its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLER
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/*
 * $Id: print.c,v 1.43 1998/06/28 18:13:25 onoe Exp $
 */

#include "mgp.h"

static u_int align = AL_LEFT;
static u_char *curprefix = NULL;
static u_char *tabprefix = NULL;
static u_int lineheight;
static u_int linewidth;

static u_int pos_y;

static int tmpopen = 0;

#define A4_WIDTH	760
#define A4_HEIGHT	590
#define A4_XMARGIN	10
#define A4_YMARGIN	5
#define PRINT_ASCIIFONT		"Times-Roman"
#define PRINT_ASCIITHICKFONT	"Helvetica"
#define PRINT_KANJIFONT		"Ryumin-Light-H"
#define PRINT_KANJITHICKFONT	"GothicBBB-Medium-H"

#define PSSCALEF	1.25
#define PSSCALE		(100.0 / 80.0)
#define PSBASE		(80.0 / 100.0)

static FILE *fp;
static u_int reverse;
static u_char outputfile[MAXVALLEN];
static int colorps = 0;
static u_int curlinenum;
static u_int curpagenum;

static struct imagepool {
	struct ctrl *image;
	int xoffset;
	int xsiz;
	int ysiz;
	int eps;
} imagepool[256];		/*enough?*/
static int nimagepool = 0;
static struct textpool {
	int pfx;
	int xoffset;
	int xsiz;
	int font;
	int size;
	char *text;
	u_long fore;
	u_long back;
} textpool[1024];		/*enough?*/
static int ntextpool = 0;

u_long fore;
u_long back;

/* the order must be the same for the following two decls. */
#define ASCII	0
#define KANJI	1
#define NOPE	-1
static int print_font_bold[] = {
	0, 0, 0, 0,
};
static char *print_font_string[] = {
	PRINT_ASCIIFONT, PRINT_ASCIITHICKFONT,
	PRINT_KANJIFONT, PRINT_KANJITHICKFONT,
	NULL,
};
static char *print_font_abbrev[] = {
	"ASC", "ASCT", "KNJ", "KNJT", NULL,
};

/* XXX: width for times-roman */
static int cwidth[] = {
	 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	25, 33, 41, 50, 50, 83, 78, 33, 33, 33, 50, 56, 25, 33, 25, 28,
	50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 28, 28, 56, 56, 56, 44,
	92, 72, 67, 67, 72, 61, 56, 72, 72, 33, 39, 72, 61, 89, 72, 72,
	56, 72, 67, 56, 61, 72, 72, 94, 72, 72, 61, 33, 28, 33, 47, 50,
	33, 44, 50, 44, 50, 44, 33, 50, 50, 27, 27, 50, 28, 78, 50, 50,
	50, 50, 33, 39, 28, 50, 50, 72, 50, 50, 44, 48, 20, 48, 54,
};

static u_short kinsokutable[] = {
	0x2121, 0x2122, 0x2123, 0x2124, 0x2125, 0x2126, 0x2127, 0x2128,
	0x2129, 0x212a, 0x212b, 0x212c, 0x212d, 0x212e, 0x212f, 0x2130,
	0x2133, 0x2134, 0x2135, 0x2136, 0x213c, 0x2147, 0x2149, 0x214b,
	0x214d, 0x214f, 0x2151, 0x2153, 0x2155, 0x2157, 0x2159, 0x216b,
	0x2242, 0x2244, 0
};

static void print_out __P((void));
static void print_page __P((u_int));
static void print_init __P((void));
static void print_pageinit __P((u_int));
static void print_pagebackground __P((void));
static void process_direc_print __P((struct ctrl *));
static void line_start __P((void));
static void line_end __P((void));
static void icon_output __P((struct textpool *));
static void line_skip __P((int, int));
static void print_bar __P((struct ctrl *));
static void print_icon __P((struct ctrl *));
static void print_line __P((u_char *, int));
static int iskinsokuchar __P((u_short));
static void print_string __P((u_char *, int));
static void print_fragment __P((u_char *, u_int, int, int));
static void text_remember __P((char *, int, int, int, int));
static void icon_remember __P((int, int, int, u_long));
static void image_remember __P((struct ctrl *, struct imagepool *));
static void image_load_print __P((char *, int, int, int, int, int));
static void print_usage __P((char *));
static int set_position __P((void));
static void setpencolor __P((u_long));
static void print_full_image __P((XImage *, Visual *));
static void print_256_image __P((XImage *));
static void print_eps __P((FILE *, char *, int, int, int, int, int));

/* image/ *.[ch] */
extern char *Paths[];
extern int NumPaths;
extern char *expandPath __P((char *));

extern int findImage __P((char *, char *));

static void
print_out()
{
	u_int	i;
	u_int	width, height;

	width  = window_width;
	height = window_height;
	window_width  = A4_WIDTH;
	window_height = A4_HEIGHT;

	char_size = window_height * DEFAULT_CHARSIZE / 100;

	if (outputfile[0]) {
		if ((fp = fopen(outputfile, "w")) == NULL) {
			perror("fopen");
			exit(-1);
		}
	} else
		fp = stdout;

	print_init();
	for (i = 1; i <= maxpage; i ++)
		print_page(i);

	fclose(fp);

	window_width  = width;
	window_height = height;
}

static void
print_page(page)
	u_int page;
{
	u_int line;
	struct ctrl *cp;

	pos_y = 0;
	print_pageinit(curpagenum);

	for (line = 0;
	     line <= page_attribute[page].pg_linenum;
	     line++) {
		curlinenum = line;
		curpagenum = page;
		/* process normal control */
		for (cp = page_control[page][line]; cp; cp = cp->ct_next)
			process_direc_print(cp);
	}

	fprintf(fp, "showpage\n\n");
}

static void
print_init()
{
	char **p;
	char **q;

	fprintf(fp, "%%!PS-Adobe-2.0 EPSF-2.0\n");
	fprintf(fp, "%%%%Creater: mgp2ps\n");
	fprintf(fp, "%%%%Title: mgp2ps output file\n");
	fprintf(fp, "%%%%Pages: %d\n", maxpage);
	fprintf(fp, "%%%%BoundingBox: 0 0 612 792\n");
	fprintf(fp, "%%%%DocumentPaperSizes: A4\n");
	fprintf(fp, "%%%%Orientation: Landscape\n");
	fprintf(fp, "%%%%EndComments\n");

	/* define constants */
	fprintf(fp, "/XMARGIN %d def /YMARGIN %d def "
		"/WIDTH %d def /HEIGHT %d def\n",
		A4_XMARGIN, A4_YMARGIN, window_width, window_height);
	fprintf(fp, "/XBODY WIDTH XMARGIN 2 mul sub def\n");

	/* define writebox */
	fprintf(fp, "/writebox { XMARGIN YMARGIN -1 mul moveto ");
	fprintf(fp, "0 HEIGHT -1 mul rlineto ");
	fprintf(fp, "WIDTH 0 rlineto ");
	fprintf(fp, "0 HEIGHT rlineto ");
	fprintf(fp, "WIDTH -1 mul 0 rlineto stroke } def\n");

	/* define writeboxfill */
	fprintf(fp, "/writeboxfill { newpath XMARGIN YMARGIN -1 mul moveto ");
	fprintf(fp, "0 HEIGHT -1 mul rlineto ");
	fprintf(fp, "WIDTH 0 rlineto ");
	fprintf(fp, "0 HEIGHT rlineto ");
	fprintf(fp, "WIDTH -1 mul 0 rlineto closepath eofill stroke } def\n");

	/* show with line wrapping */
	fprintf(fp, "/initcharsize {/charsize 0 def} def initcharsize\n");
	fprintf(fp, "/setcharsize {\n");
	fprintf(fp, "  dup charsize gt {dup /charsize exch def}if pop\n");
	fprintf(fp, "} def\n");

	/* show with line wrapping */
	fprintf(fp, "/showwithwrap {\n");
	fprintf(fp, "  1 copy stringwidth pop currentpoint pop add\n");
	fprintf(fp, "  XMARGIN XBODY add gt\n");
	fprintf(fp, "  {\n");
	fprintf(fp, "    /ypos ypos charsize sub def\n");
	fprintf(fp, "    xpos fillzero add ypos charsize sub moveto\n");
	fprintf(fp, "  }if\n");
	fprintf(fp, "  show\n");
	fprintf(fp, "} def\n");

	/* define font calling sequence */
	for (p = print_font_string, q = print_font_abbrev; *p && *q; p++, q++) {
		fprintf(fp, "/%s {/%s findfont exch scalefont setfont} def\n",
			*q, *p);
	}

	/* a spell for EPS */
	fprintf(fp, "%%\n");
	fprintf(fp, "/BeginEPSF {%%def\n");
	fprintf(fp, "  /b4_Inc_state save def\n");
	fprintf(fp, "  /dict_count countdictstack def\n");
	fprintf(fp, "  /op_count count 1 sub def\n");
	fprintf(fp, "  userdict begin\n");
	fprintf(fp, "  /showpage {}def\n");
	fprintf(fp, "  0 setgray 0 setlinecap\n");
	fprintf(fp, "  1 setlinewidth 0 setlinejoin\n");
	fprintf(fp, "  10 setmiterlimit [] 0 setdash\n");
	fprintf(fp, "  newpath\n");
	fprintf(fp, "  /languagelevel where\n");
	fprintf(fp, "  {pop languagelevel\n");
	fprintf(fp, "  1 ne\n");
	fprintf(fp, "    {false setstrokeadjust\n");
	fprintf(fp, "     false setoverprint\n");
	fprintf(fp, "    }if\n");
	fprintf(fp, "  }if\n");
	fprintf(fp, "}bind def\n");
	fprintf(fp, "%%\n");
	fprintf(fp, "/EndEPSF {%%def\n");
	fprintf(fp, "  count op_count sub {pop}repeat\n");
	fprintf(fp, "  countdictstack\n");
	fprintf(fp, "  dict_count sub {end}repeat\n");
	fprintf(fp, "  b4_Inc_state restore\n");
	fprintf(fp, "}bind def\n");
	fprintf(fp, "%%\n");
}

static void
print_pageinit(page)
	u_int page;
{
	fprintf(fp, "%%%%Page: %d %d\n", page, page);
	fprintf(fp, "/ypos %d def\n", - A4_YMARGIN);
	fprintf(fp, "initcharsize\n");
}

static void
print_pagebackground()
{
	if (!colorps)
		fprintf(fp, "90 rotate newpath writebox\n");
	else {
		setpencolor(back);
		fprintf(fp, "90 rotate newpath writeboxfill\n");
	}
}

static void
process_direc_print(cp)
	struct ctrl *cp;
{
	switch(cp->ct_op) {
	case CTL_SIZE:
		char_size = window_height * cp->cti_value / 100;
		fprintf(fp, "%d setcharsize\n", char_size);
		break;

#ifdef VFLIB
	case CTL_VFONT:
		print_font_bold[KANJI] = strstr(cp->ctc_value, "goth") ? 1 : 0;
		break;
#endif /*VFLIB*/

	case CTL_XFONT:
		print_font_bold[ASCII] =
			strstr(cp->ctc_value, "helvetica") ? 1 : 0;
		break;

	case CTL_VGAP:
		vert_gap = cp->cti_value;
		break;

	case CTL_HGAP:
		horiz_gap = cp->cti_value;
		break;

	case CTL_GAP:
		vert_gap = horiz_gap = cp->cti_value;
		break;

	case CTL_CENTER:
		align = AL_CENTER;
		break;

	case CTL_LEFT:
		align = AL_LEFT;
		break;

	case CTL_LEFTFILL:
		align = AL_LEFTFILL0;
		break;

	case CTL_RIGHT:
		align = AL_RIGHT;
		break;

	case CTL_IMAGE:
		image_remember(cp, &imagepool[nimagepool]);
		line_skip(imagepool[nimagepool].xsiz,
			imagepool[nimagepool].ysiz);
		nimagepool++;
		break;

	case CTL_BAR:
		print_bar(cp);
		break;

	case CTL_PREFIX:
		curprefix = cp->ctc_value;
		break;

	case CTL_TABPREFIX:
		tabprefix = cp->ctc_value;
		break;

	case CTL_PREFIXPOS:
		if (tabprefix)
			print_line(tabprefix, 1);
		else if (curprefix)
			print_line(curprefix, 1);
		break;

	case CTL_TEXT:
	    {
		if (!cp->ctc_value)
			break;
		print_line(cp->ctc_value, 0);
		if (align == AL_LEFTFILL0)
			align = AL_LEFTFILL1;
		break;
	    }

	case CTL_ICON:
		print_icon(cp);
		break;

	case CTL_LINESTART:
		line_start();
		break;

	case CTL_LINEEND:
		line_end();
		if (align == AL_LEFTFILL1)
			align = AL_LEFTFILL0;
		break;

	case CTL_FORE:
		fore = cp->ctl_value;
		break;

	case CTL_BACK:
		back = cp->ctl_value;
		break;

	default:
		break;
	}
}

static void
line_start()
{
	linewidth = 0;
	nimagepool = 0;
	ntextpool = 0;

	if (curlinenum == 0)
		print_pagebackground();
}

static void
line_end()
{
	int xpos;
	int ypos;
	int i;
	char *t;
	int fill0;

	/*
	 * compute default guess, by our side...
	 * the value is not confirmed by PostScript engine, so there will be
	 * a huge error.
	 */
	switch (align) {
	case AL_CENTER:
		xpos = (window_width - linewidth)/ 2;
		break;
	case AL_LEFT:
	case AL_LEFTFILL0:
	case AL_LEFTFILL1:
		xpos = 0;
		break;
	case AL_RIGHT:
		xpos = window_width - linewidth;
		break;
	default:
		xpos = 0;
		break;
	}
	xpos += A4_XMARGIN;
	ypos = - (pos_y + VERT_GAP(char_size)/2) - A4_YMARGIN - char_size;

	/*
	 * line output starts here
	 */
	if (mgp_flag & FL_VERBOSE)
		fprintf(fp, "%% a line starts here\n");

	/*
	 * if we have nothing on the line, skip it
	 */
	if (!ntextpool && !nimagepool)
		goto done;

	/*
	 * push strings in reverse order.
	 */
#if 0
	fprintf(fp, "/ypos %d def\n", ypos);
#endif
#if 0
	fprintf(fp, "/xpos %d def /ypos %d def\n", xpos, ypos);

	/* just for safety - make rmoveto happy */
	if (ntextpool || nimagepool)
		fprintf(fp, "%d %d moveto\n", xpos, ypos);
#endif

	fprintf(fp, "initcharsize\n");
	fprintf(fp, "0 %% sentinel for text width computation\n");

	for (i = ntextpool - 1; 0 <= i; i--) {
		if (!textpool[i].text) {
			fprintf(fp, "%d add\n", textpool[i].xsiz);
			continue;
		}
		fprintf(fp, "%d setcharsize %d %s %s 1 copy stringwidth pop 3 2 roll add\n",
			textpool[i].size, textpool[i].size,
			print_font_abbrev[textpool[i].font],
			textpool[i].text);
	}
	if (mgp_flag & FL_VERBOSE)
		fprintf(fp, "%% stack should have str3 str2 str1 width\n");
	if (align == AL_CENTER) {
		if (mgp_flag & FL_VERBOSE)
			fprintf(fp, "%% centering\n");
		fprintf(fp, "XBODY exch sub 2 div XMARGIN add "
			"/xpos exch def\n");
	} else if (align == AL_RIGHT) {
		if (mgp_flag & FL_VERBOSE)
			fprintf(fp, "%% rightfill\n");
		fprintf(fp, "XBODY exch sub XMARGIN add /xpos exch def\n");
	} else {
		if (mgp_flag & FL_VERBOSE)
			fprintf(fp, "%% leftfill\n");
		fprintf(fp, "pop /xpos XMARGIN def\n");
	}
	fprintf(fp, "xpos ypos charsize sub moveto\n");
	fill0 = 0;
	for (i = 0; i < ntextpool; i++) {
		if (colorps
		 && (i == 0
		  || (textpool[i - 1].fore != textpool[i].fore)
		  || (textpool[i - 1].back != textpool[i].back))) {
			setpencolor(textpool[i].fore);
		}

		if (!textpool[i].text) {
			icon_output(&textpool[i]);
			continue;
		}

		if (!fill0
		&&  !textpool[i].pfx
		&&  (align == AL_LEFTFILL0 || align == AL_LEFTFILL1)) {
			fprintf(fp, "currentpoint pop /fillzero exch def\n");
			fill0 = 1;
		}
		fprintf(fp, "%d %s ", textpool[i].size,
			print_font_abbrev[textpool[i].font]);
		if (fill0)
			fprintf(fp, "showwithwrap\n");
		else
			fprintf(fp, "show\n");
	}
	ntextpool = 0;

    {
	struct ctrl *cp1;

	for (i = 0; i < nimagepool; i++) {
		if (mgp_flag & FL_VERBOSE)
			fprintf(fp, "%% emit the content of imagepool\n");
		cp1 = imagepool[i].image;
		if (!cp1)
			continue;
		fprintf(fp, "xpos %d add ypos moveto\n",
			imagepool[i].xoffset);
		image_load_print(cp1->ctm_fname, cp1->ctm_numcolor,
			cp1->ctm_ximagesize, cp1->ctm_yimagesize, 0,
			cp1->ctm_zoomflag);
		fprintf(fp, "%d setcharsize\n", imagepool[i].ysiz);	/*XXX*/
	}
	nimagepool = 0;
    }

done:
	/* see definition of VERT_GAP */
#if 1
	fprintf(fp, "%d 100 div charsize mul charsize add\n", vert_gap);
	fprintf(fp, "ypos exch sub /ypos exch def\n");
#else
	fprintf(fp, "/ypos ypos charsize sub def\n");
#endif
	pos_y += VERT_STEP(char_size);
	tabprefix = NULL;
}

static void
icon_output(tp)
	struct textpool *tp;
{
	int isize = tp->size;
	int csize = tp->xsiz;
	int ixoff, iyoff;

	iyoff = (csize - isize) / 3;	/*XXX*/
	ixoff = tp->xoffset + iyoff;

	switch (tp->font) {
	case 0:
		/* XXX: image is not supported yet */
		break;
	case 1:
		fprintf(fp, "currentpoint ");
		if (colorps)
			fprintf(fp, "currentpoint newpath moveto ");
		fprintf(fp, "%d %d rmoveto ", ixoff, iyoff);
		fprintf(fp, "0 %d rlineto ", isize);
		fprintf(fp, "%d 0 rlineto ", isize);
		fprintf(fp, "0 %d rlineto ", -isize);
		fprintf(fp, "%d 0 rlineto ", -isize);
		if (colorps)
			fprintf(fp, "closepath eofill ");
		fprintf(fp, "stroke moveto\n");
		fprintf(fp, "%d 0 rmoveto\n", ixoff * 2 + isize);
		break;
	case 2:
		fprintf(fp, "currentpoint ");
		if (colorps)
			fprintf(fp, "currentpoint newpath moveto ");
		fprintf(fp, "%d %d rmoveto ", ixoff + isize, iyoff + isize/2);
		fprintf(fp, "currentpoint exch %d add exch %d 0 360 arc ",
			-isize/2, isize/2);
		if (colorps)
			fprintf(fp, "closepath eofill ");
		fprintf(fp, "stroke moveto\n");
		fprintf(fp, "%d 0 rmoveto\n", ixoff * 2 + isize);
		break;
	case 3:
		fprintf(fp, "currentpoint ");
		if (colorps)
			fprintf(fp, "currentpoint newpath moveto ");
		fprintf(fp, "%d %d rmoveto ", ixoff, iyoff);
		fprintf(fp, "%d 0 rlineto ", isize);
		fprintf(fp, "%d %d rlineto ", -isize/2, isize);
		fprintf(fp, "%d %d rlineto ", -isize/2, -isize);
		if (colorps)
			fprintf(fp, "closepath eofill ");
		fprintf(fp, "stroke moveto\n");
		fprintf(fp, "%d 0 rmoveto\n", ixoff * 2 + isize);
		break;
	case 4:
		fprintf(fp, "currentpoint ");
		if (colorps)
			fprintf(fp, "currentpoint newpath moveto ");
		fprintf(fp, "%d %d rmoveto ", ixoff, iyoff + isize);
		fprintf(fp, "%d 0 rlineto ", isize);
		fprintf(fp, "%d %d rlineto ", -isize/2, -isize);
		fprintf(fp, "%d %d rlineto ", -isize/2, isize);
		if (colorps)
			fprintf(fp, "closepath eofill ");
		fprintf(fp, "stroke moveto\n");
		fprintf(fp, "%d 0 rmoveto\n", ixoff * 2 + isize);
		break;
	case 5:
		fprintf(fp, "currentpoint ");
		if (colorps)
			fprintf(fp, "currentpoint newpath moveto ");
		fprintf(fp, "%d %d rmoveto ", ixoff, iyoff);
		fprintf(fp, "0 %d rlineto ", isize);
		fprintf(fp, "%d %d rlineto ", isize, -isize/2);
		fprintf(fp, "%d %d rlineto ", -isize, -isize/2);
		if (colorps)
			fprintf(fp, "closepath eofill ");
		fprintf(fp, "stroke moveto\n");
		fprintf(fp, "%d 0 rmoveto\n", ixoff * 2 + isize);
		break;
	case 6:
		fprintf(fp, "currentpoint ");
		if (colorps)
			fprintf(fp, "currentpoint newpath moveto ");
		fprintf(fp, "%d %d rmoveto ", ixoff + isize, iyoff);
		fprintf(fp, "0 %d rlineto ", isize);
		fprintf(fp, "%d %d rlineto ", -isize, -isize/2);
		fprintf(fp, "%d %d rlineto ", isize, -isize/2);
		if (colorps)
			fprintf(fp, "closepath eofill ");
		fprintf(fp, "stroke moveto\n");
		fprintf(fp, "%d 0 rmoveto\n", ixoff * 2 + isize);
		break;
	case 7:
		fprintf(fp, "currentpoint ");
		if (colorps)
			fprintf(fp, "currentpoint newpath moveto ");
		fprintf(fp, "%d %d rmoveto ", ixoff, iyoff + isize/2);
		fprintf(fp, "%d %d rlineto ",  isize/2,  isize/2);
		fprintf(fp, "%d %d rlineto ",  isize/2, -isize/2);
		fprintf(fp, "%d %d rlineto ", -isize/2, -isize/2);
		fprintf(fp, "%d %d rlineto ", -isize/2,  isize/2);
		if (colorps)
			fprintf(fp, "closepath eofill ");
		fprintf(fp, "stroke moveto\n");
		fprintf(fp, "%d 0 rmoveto\n", ixoff * 2 + isize);
		break;
	}
}

static void
line_skip(x, y)
	int x;
	int y;
{
	linewidth += x;
	lineheight = (lineheight < y) ? y : lineheight;
}

static void
print_bar(cp)
	struct ctrl *cp;
{
	struct ctrl_bar pbar;

	pbar.ct_width = cp->ctb_width * A4_HEIGHT / 1000;
	pbar.ct_start = cp->ctb_start * A4_WIDTH / 100;
	pbar.ct_length = cp->ctb_length * A4_WIDTH / 100;

	fprintf(fp, "%%bar color %d %d %d\n",
		cp->ctb_width, cp->ctb_start, cp->ctb_length);
	fprintf(fp, "XMARGIN ypos moveto\n");
	fprintf(fp, "%d 0 rmoveto\n", pbar.ct_start);
	fprintf(fp, "%d %d rlineto\n", 0, pbar.ct_width * -1);
	fprintf(fp, "%d %d rlineto\n", pbar.ct_length, 0);
	fprintf(fp, "%d %d rlineto\n", 0, pbar.ct_width);
	fprintf(fp, "%d %d rlineto stroke\n", pbar.ct_length * -1, 0);
	fprintf(fp, "/ypos ypos %d sub def\n",
		cp->ctb_width + VERT_GAP(char_size) / 2);
	fprintf(fp, "xpos ypos moveto\n");

	pos_y += cp->ctb_width + VERT_GAP(char_size) / 2;
	linewidth = 0;
}

static void
print_icon(cp)
	struct ctrl *cp;
{
	int i;
	int itype, isize;
	static struct ctl_words icon_words[] = {
		{ 1, "box", 3 },
		{ 2, "arc", 3 },
		{ 3, "delta1", 6 },
		{ 4, "delta2", 6 },
		{ 5, "delta3", 6 },
		{ 6, "delta4", 6 },
		{ 7, "dia", 3 },
		{ 0, NULL, 0 }
	};

	for (i = 0; icon_words[i].ctl_strlen != 0; i++) {
		if (!strncasecmp(cp->ctic_value, icon_words[i].ctl_string,
			strlen(cp->ctic_value)))
				break;
	}
	itype = icon_words[i].ctl_type;
	isize = char_size * cp->ctic_size / 100;
	icon_remember(itype, isize, linewidth, cp->ctic_color);
	linewidth += char_size;
}

static void
print_line(data, pfx)
	u_char *data;
	int pfx;
{
	if (data && *data)
		print_string(data, pfx);
}

static int
iskinsokuchar(code)
	u_short code;
{
	u_short *kinsoku;

	for (kinsoku = kinsokutable; *kinsoku; kinsoku++) {
		if (code == *kinsoku)
			return 1;
	}
	return 0;
}

static void
print_string(data, pfx)
	u_char *data;
	int pfx;
{
	u_char *p;
	u_char *q;
	int kanji = 0;
	u_int code2;

	p = data;

	while (*p && *p != '\n') {
		if (p[0] == 0x1b && p[1] == '$'
		 && (p[2] == 'B' || p[2] == '@')) {
			kanji = 1;
			p += 3;
			continue;
		}
		if (p[0] == 0x1b && p[1] == '('
		 && (p[2] == 'B' || p[2] == 'J')) {
			kanji = 0;
			p += 3;
			continue;
		}

#if 0
		if (!kanji && isspace(p[0])) {
			p++;
			linewidth +=
			continue;
		}
#endif

		if (kanji) {
			for (q = p + 2; 0x21 <= q[0] && q[0] <= 0x7e; q += 2) {
				code2 = q[0] * 256 + q[1];
				if (!iskinsokuchar(code2))
					break;
			}
		} else {
			q = p;
			while (*q && isprint(*q) && !isspace(*q))
				q++;
			if (q == p)
				q++;
			else {
				/* append spaces to the end of the word. */
				while (*q && isspace(*q))
					q++;
			}
		}

		print_fragment(p, q - p, kanji, pfx);

		p = q;
	}
}

static void
print_fragment(data, len, kanjimode, pfx)
	u_char *data;
	u_int len;
	int kanjimode;
	int pfx;
{
	u_char *p;
	u_char *q;
	char buf[4096];
	u_int code;
	int textstartpos = 0;
#define OPENBRACE(kanji)	((kanji) ? '<' : '(')
#define CLOSEBRACE(kanji)	((kanji) ? '>' : ')')
#define	DANGERLETTER(x)	\
	((x) == '<' || (x) == '>' || (x) == '(' || (x) == ')' \
	 || (x) == '\\' || (x) == '%')

	p = data;
	if (kanjimode) {
		q = &buf[1];
		while (len) {
			code = p[0] * 256 + p[1];
			sprintf(q, "%04x ", code);
			p += 2;
			len -= 2;
			q += 5;
		}
		buf[0] = OPENBRACE(kanjimode);
		*q++ = CLOSEBRACE(kanjimode);
		*q = '\0';
		text_remember(buf, KANJI, char_size, textstartpos, pfx);
	} else {
		while (len) {
			if (DANGERLETTER(p[0])) {
				sprintf(buf, "<%02x>", p[0]);
				text_remember(buf, ASCII, char_size,
					textstartpos, pfx);
				len--;
				p++;
				continue;
			}
			q = p;
			while (q - p < len && !DANGERLETTER(q[0]))
				q++;
			buf[0] = OPENBRACE(kanjimode);
			memcpy(&buf[1], p, q - p);
			buf[q - p + 1] = CLOSEBRACE(kanjimode);
			buf[q - p + 2] = '\0';
			text_remember(buf, ASCII, char_size, textstartpos, pfx);
			len -= (q - p);
			p = q;
		}
	}
}

static char *
checkeps(fname)
	char *fname;
{
	static char epsfile[MAXPATHLEN];
	static char fullname[MAXPATHLEN];
	char *p;
	FILE *epsfp;

	strcpy(epsfile, fname);

	p = strrchr(epsfile, '.');
	if (p)
		*p = '\0';
	strcat(epsfile, ".eps");

	fullname[0] = '\0';
	if (findImage(epsfile, fullname) == 0)
		return fullname;
	else
		return NULL;
}

static void
text_remember(text, ctype, fontsize, offset, pfx)
	char *text;
	int ctype;
	int fontsize;
	int offset;
	int pfx;
{
	/*XXX*/
	if (strcmp(text, "()") == 0)
		return;
	textpool[ntextpool].pfx = pfx;
	textpool[ntextpool].xoffset = offset;
	textpool[ntextpool].xsiz = linewidth - offset;
	textpool[ntextpool].size = fontsize;
	textpool[ntextpool].font = ctype * 2 + (print_font_bold[ctype] ? 1 : 0);
	textpool[ntextpool].text = strdup(text);
	textpool[ntextpool].fore = fore;
	textpool[ntextpool].back = back;
	ntextpool++;
}

static void
icon_remember(icon, fontsize, offset, color)
	int icon;
	int fontsize;
	int offset;
	u_long color;
{
	textpool[ntextpool].pfx = 1;
	textpool[ntextpool].xoffset = offset;
	textpool[ntextpool].xsiz = char_size;
	textpool[ntextpool].size = fontsize;
	textpool[ntextpool].font = icon;
	textpool[ntextpool].text = NULL;
	textpool[ntextpool].fore = color;
	textpool[ntextpool].back = back;	/*XXX*/
	ntextpool++;
}

static void
image_remember(cp, pool)
	struct ctrl *cp;
	struct imagepool *pool;
{
	char *epsname;
	char buf[BUFSIZ];
	Image *myimage;

	if (epsname = checkeps(cp->ctm_fname)) {
		int x1, y1, x2, y2, width, height, swidth, sheight;

		if (ps_boundingbox(epsname, &x1, &y1, &x2, &y2) < 0)
			goto noneps;

		width = x2 - x1;
		height = y2 - y1;
		swidth = (int)(width * PSSCALE);
		sheight = (int)(height * PSSCALE);

		pool->xsiz = width;
		pool->ysiz = height;
		pool->xoffset = linewidth;
		pool->image = cp;
		return;
	}

noneps:
	myimage = loadImage(cp->ctm_fname, mgp_flag & FL_VERBOSE);
	if (!myimage) {
		fprintf(stderr, "failed to open %s\n", cp->ctm_fname);
		exit(1);
	}
	pool->xsiz = myimage->width;
	pool->ysiz = myimage->height;
	freeImage(myimage);
	pool->xoffset = linewidth;
	pool->image = cp;
}

static void
image_load_print(filename, numcolor, ximagesize, yimagesize, backflag, zoomflag)
	char *filename;
	int numcolor;
	int ximagesize;
	int yimagesize;
	int backflag;
	int zoomflag;
{
	Image *image, *myimage;
	Pixmap mypixmap;
	XImageInfo *ximageinfo;
	XImage	*print_image;
	int width, height;
	int xzoomrate, yzoomrate;
	int verbose = mgp_flag & FL_VERBOSE;
	static Cursor curs;
	u_int	print_width, print_height;

	int xpos;

	if (zoomflag == 2)
		ximagesize = yimagesize = 0;

	/* hook for eps */
    {
	char *p;
	FILE *epsfp;

	p = checkeps(filename);
	if (p) {
		epsfp = fopen(p, "r");
		print_eps(epsfp, p, numcolor, ximagesize, yimagesize,
			backflag, zoomflag);
		fclose(epsfp);
		return;
	}
    }

	if (!curs)
		curs = XCreateFontCursor(display, XC_watch);
	XDefineCursor(display, window, curs); XFlush(display);

	screen = DefaultScreen(display);

	if ((myimage = loadImage(filename, verbose)) == NULL)
		exit(-1);	/* fail to load image data */
	width = myimage->width;
	height = myimage->height;

	if (numcolor)
		myimage = reduce(myimage, numcolor, verbose);

	if (ximagesize != 0 && yimagesize != 0) {
		if (!zoomflag) {
			xzoomrate = window_width * ximagesize / width;
			yzoomrate = window_height * yimagesize / height;
		} else {
			xzoomrate = ximagesize;
			yzoomrate = yimagesize;
		}
		image = myimage;
		myimage = zoom(image, xzoomrate, yzoomrate, verbose);
		freeImage(image);
	} else
		xzoomrate = yzoomrate = 100;

	if (! (ximageinfo= imageToXImage(display, screen, visual, depth,
			myimage, 0, 0, 0, verbose))) {
		fprintf(stderr, "Cannot convert Image to XImage\n");
		exit(1);
	}

	mypixmap = ximageToPixmap(display, RootWindow(display, 0), ximageinfo);
	if (!mypixmap) {
		fprintf(stderr, "Cannot create image in server\n");
		exit(1);
	}

	switch (align) {
	case AL_LEFT:
	case AL_LEFTFILL0:
	case AL_LEFTFILL1:
		xpos = 0;
		break;
	case AL_CENTER:
		xpos = (window_width - width * xzoomrate /100) / 2;
		break;
	case AL_RIGHT:
		xpos = window_width - width * xzoomrate /100;
		break;
	}

	print_width = myimage->width;
	print_height = myimage->height;

	print_image =  XGetImage(display, mypixmap, 0, 0,
					print_width, print_height,
					AllPlanes, ZPixmap);

	fprintf(fp, "gsave\n");
	fprintf(fp, "%d %d translate\n", xpos + A4_XMARGIN,
		-(pos_y + print_height + A4_YMARGIN));
	fprintf(fp, "%d %d scale\n", print_width, print_height);
	fprintf(fp, "%d %d 8\n", print_width, print_height);
	fprintf(fp, "[%d 0 0 %d 0 %d]\n", print_width,
			-1 * print_height, print_height);
	fprintf(fp, "{currentfile\n");
	fprintf(fp, "%d string readhexstring pop}\n", print_width * 3);
	fprintf(fp, "false 3\n");
	fprintf(fp, "colorimage\n");

	if (print_image->bits_per_pixel == 8) {
			print_256_image(print_image);
	 }else {
		if (print_image->bits_per_pixel == 16) {
			print_full_image(print_image, visual);
		} else {
			fprintf(stderr, "Sorry unsupported visual %d\n",
				print_image->bits_per_pixel);
			exit(-1);
		}
	}

	fprintf(fp, "grestore\n");

	XFreePixmap(display, mypixmap);
	freeXImage(myimage, ximageinfo);
	freeImage(myimage);
	XUndefineCursor(display, window); XFlush(display);

	pos_y += print_height - char_size;
}

static void
print_usage(name)
	char	*name;
{
	fprintf(stderr, "Usage: %s [-crV] [-f psfile] mgpfile\n", name);
	exit(0);
}

static int
set_position()
{
	u_int   x;

	switch(align) {
	case AL_CENTER:
		x = (window_width - linewidth)/ 2;
		break;

	case AL_LEFT:
	case AL_LEFTFILL0:
	case AL_LEFTFILL1:
		x = 0;
		break;

	case AL_RIGHT:
		x = window_width - linewidth;
		break;

	default:
		x = 0;
		break;
	}

	return x;
}

static void
setpencolor(c)
	u_long c;
{
	/*
	 * update the color of the pen.
	 * XXX background is ignored at this moment
	 */
	int mask;
	XColor color;

	color.pixel = c;
	XQueryColor(display, colormap, &color);
#if 0
	if (color.red == color.green && color.green == color.blue) {
		/* XXX: invert white-to-gray colors */
		color.red = 65535 - color.red;
		color.green = 65535 - color.green;
		color.blue = 65535 - color.blue;
	}
#endif
	fprintf(fp, "%f %f %f setrgbcolor %% #%08x depth %d\n",
		color.red / 65535.0,
		color.green / 65535.0,
		color.blue / 65535.0,
		color.pixel, depth);
}

static void
print_full_image(print_image, visual)
	XImage	*print_image;
	Visual	*visual;
{
	u_int	x, y;
	u_int	r, g, b;
	u_int	count = 0;
	u_int	width, height, byte_width;
	u_short	*data;

	data = (u_short *)print_image->data;
	width = print_image->width;
	height = print_image->height;
	byte_width = print_image->bytes_per_line / sizeof(u_short);

	for (y = 0; y < height; y ++) {
		for (x = 0; x < width; x ++) {
			r = ((data[x + y * byte_width] & visual->red_mask) >> 11) & 0xff;
			g = ((data[x + y * byte_width] & visual->green_mask) >> 5) & 0xff;
			b = (data[x + y * byte_width] & visual->blue_mask) & 0xff;

			if (reverse)
				fprintf(fp, "%02x%02x%02x",
					~(r << 3) & 0xff, ~(g << 2) & 0xff, ~(b << 3) & 0xff);
			else
				fprintf(fp, "%02x%02x%02x", r << 3, g << 2, b << 3);

			count ++;
			if (count == 10) {
				count = 0;
				fprintf(fp, "\n");
			}
		}
	}
}

static void
print_256_image(print_image)
	XImage	*print_image;
{
	u_int	x, y;
	u_int	count = 0;
	u_char	*data;
	u_int	width, height, byte_width;

	data = (u_char *)print_image->data;
	width = print_image->width;
	height = print_image->height;
	byte_width = print_image->bytes_per_line / sizeof(u_char);

	for (y = 0; y < height; y ++) {
		for (x = 0; x < width; x ++) {
			if (reverse)
				fprintf(fp, "%02x%02x%02x",
					~data[x + y * byte_width] & 0xff,
					~data[x + y * byte_width] & 0xff,
					~data[x + y * byte_width] & 0xff);
			else
				fprintf(fp, "%02x%02x%02x",
					data[x + y * byte_width],
					data[x + y * byte_width],
					data[x + y * byte_width]);

			count ++;
			if (count == 10) {
				count = 0;
				fprintf(fp, "\n");
			}
		}
	}
}

static void
print_eps(epsfp, filename, numcolor, ximagesize, yimagesize, backflag, zoomflag)
	FILE *epsfp;
	char *filename;
	int numcolor;
	int ximagesize;
	int yimagesize;
	int backflag;
	int zoomflag;
{
	char line1[BUFSIZ];
	char line2[BUFSIZ];
	char line[BUFSIZ];
	int x1, y1, x2, y2, height, width, sheight, swidth;
	int xpos;

	/*
	 * mgp scale is NOT supported.
	 *
	 * PS scale code is for visibility when printed.
	 * Kazu's epsfile should be enlarged 1.2 times.
	 * And this is a challenge to understand PS's scale...
	 */
	if (fgets(line1, sizeof(line1), epsfp) == NULL) {
		fprintf(stderr, "no first line in %s.\n", filename);
		exit(1);
	}
	if (strncmp(line1, "%!", 2) != 0) {
		fprintf(stderr, "non eps file %s used as eps.\n", filename);
		exit(1);
	}

	for (;;) {
		if (fgets(line2, sizeof(line2), epsfp) == NULL || line2[0] != '%') {
			fprintf(stderr, "no bounding box in %s.\n", filename);
			exit(1);
		}
		if (sscanf(line2, "%%%%BoundingBox: %d %d %d %d", &x1, &y1, &x2, &y2) == 4)
			break;
	}

	width = x2 - x1;
	height = y2 - y1;
	swidth = (int)(width * PSSCALE);
	sheight = (int)(height * PSSCALE);

	switch (align) {
	case AL_LEFT:
	case AL_LEFTFILL0:
	case AL_LEFTFILL1:
		xpos = 0;
		break;
	case AL_CENTER:
		xpos = (window_width - swidth) / 2;
		break;
	case AL_RIGHT:
		xpos = window_width - swidth;
		break;
	}

	fprintf(fp, "BeginEPSF\n");
	fprintf(fp, "%f %f scale\n", PSSCALEF, PSSCALEF);
	fprintf(fp, "%d %d translate\n",
		(x1 - (int)(A4_XMARGIN * PSBASE)) * -1,
		(y2 + (int)(A4_YMARGIN * PSBASE)) * -1);
	fprintf(fp, "%d %d translate\n",
		(int)(xpos * PSBASE),
		(int)(pos_y * PSBASE * -1));
	fprintf(fp, "%%BeginDocument: %s\n", filename);
	fputs(line1, fp);
	fputs(line2, fp);
	while (fgets(line, sizeof(line), epsfp)) {
		if (strncmp(line, "%%PageBoundingBox:", 18) == 0
		||  strncmp(line, "%%Trailer", 9) == 0)
			line[1] = '#';
		fputs(line, fp);
	}
	fprintf(fp, "%%EndDocument\n");
	fprintf(fp, "EndEPSF\n");

	pos_y += sheight + VERT_GAP(char_size);
}

/*------------------------------------------------------------*/

int
main(argc, argv)
	int argc;
	char **argv;
{
	u_int opt;
	extern char *optarg;
	extern int optind;
	u_char *progname;
	static char pathbuf[MAXPATHLEN];

	progname = argv[0];

	while ((opt = getopt(argc, argv, "cVrf:")) != -1) {
		switch (opt) {
		case 'c':
			colorps++;
			break;

		case 'f':
			strcpy(outputfile, optarg);
			break;

		case 'r' :
			reverse = 1;
			break;

		case 'V' :
			mgp_flag |= FL_VERBOSE;
			break;

		default:
			print_usage(progname);
			/*NOTREACHED*/
		}
	}

	argc -= optind;
	argv += optind;

	if (argc != 1) {
		print_usage(progname);
		/*NOTREACHED*/
	}

	mgp_fname = argv[0];

	/* setting up the search path. */
    {
	char *p;
	loadPathsAndExts();
	strncpy(pathbuf, mgp_fname, sizeof(pathbuf));
	if (p = rindex(pathbuf, '/')) {
		*p = '\0';
        	Paths[NumPaths++]= expandPath(pathbuf);
	}
    }

	init_win1(NULL);
	load_file(mgp_fname);
	init_win2();
	fore = 0UL;
	back = (1UL << depth) - 1;
	print_out();

	exit(0);
}
